/* 
 * delta_stream.c -- Internal text delta representation
 * 
 * ================================================================
 * Copyright (c) 2000 Collab.Net.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * 3. The end-user documentation included with the redistribution, if
 * any, must include the following acknowlegement: "This product includes
 * software developed by Collab.Net (http://www.Collab.Net/)."
 * Alternately, this acknowlegement may appear in the software itself, if
 * and wherever such third-party acknowlegements normally appear.
 * 
 * 4. The hosted project names must not be used to endorse or promote
 * products derived from this software without prior written
 * permission. For written permission, please contact info@collab.net.
 * 
 * 5. Products derived from this software may not use the "Tigris" name
 * nor may "Tigris" appear in their names without prior written
 * permission of Collab.Net.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL COLLAB.NET OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ====================================================================
 * 
 * This software may consist of voluntary contributions made by many
 * individuals on behalf of Collab.Net.
 */



#include <assert.h>
#include <stdlib.h>
#include "delta_stream.h"



/* Allocate and initalize delta window. */
svn_error_t*
svn_tdelta__init_window (svn_delta_window_t** window,
                         svn_delta_stream_t* stream)
{
  apr_pool_t* const pool = apr_make_sub_pool (stream->pool, NULL);
  if (!pool)
    return svn_create_error (APR_ENOMEM, 0,
                             "creating pool for delta window",
                             NULL, stream->pool);
  (*window) = apr_palloc (pool, sizeof (**window));
  if (!*window)
    return svn_create_error (APR_ENOMEM, 0,
                             "allocating delta window",
                             NULL, pool);

  (*window)->num_ops = 0;
  (*window)->ops_size = 0;
  (*window)->ops = NULL;
  (*window)->new = svn_string_create ("", pool);
  (*window)->pool = pool;
  return NULL;
}



/* Insert op into delta window. */
svn_error_t*
svn_tdelta__insert_op (svn_delta_window_t* window,
                       int opcode,
                       apr_off_t offset,
                       apr_off_t length,
                       char const* new_data)
{
  svn_delta_op_t* op;

  /* Create space for the new op. */
  if (window->num_ops == window->ops_size)
    {
      svn_delta_op_t* const old_ops = window->ops;
      int const new_ops_size = (window->ops_size == 0
                                ? 16 : 2 * window->ops_size);
      window->ops =
        apr_palloc (window->pool, new_ops_size * sizeof (*window->ops));
      if (!window->ops)
        return svn_create_error (APR_ENOMEM, 0,
                                 "allocating space for ops in delta windww",
                                 NULL, window->pool);

      /* Copy any existing ops into the new array */
      if (old_ops)
        memcpy (window->ops, old_ops,
                window->ops_size * sizeof (*window->ops));

      window->ops_size = new_ops_size;
    }

  /* Insert the op. svn_delta_source and svn_delta_target are
     just inserted. For svn_delta_new, the new data must be
     copied into the window. */
  op = &window->ops[window->num_ops];
  switch (opcode)
    {
    case svn_delta_source:
    case svn_delta_target:
      op->op = opcode;
      op->offset = offset;
      op->length = length;
      break;
    case svn_delta_new:
      op->op = opcode;
      op->offset = window->new->len;
      op->length = length;
      svn_string_appendbytes (window->new, new_data, length, window->pool);
      break;
    default:
      assert (!"unknown delta op.");
    }

  ++window->num_ops;
  return NULL;
}



/* The default window size. Smoeday we'll have to find a better default ... */
size_t svn_tdelta__window_size = 16384;



svn_error_t*
svn_text_delta (svn_delta_read_fn_t* source_fn,
                void* source_data,
                svn_delta_read_fn_t* target_fn,
                void* target_data,
                svn_delta_stream_t** stream)
{
  apr_pool_t* const pool = apr_make_sub_pool (NULL, NULL);
  if (!pool)
    return svn_create_error (APR_ENOMEM, 0,
                             "creating pool for delta stream",
                             NULL, NULL);
  
  *stream = apr_palloc (pool, sizeof (**stream));
  if (!*stream)
    return svn_create_error (APR_ENOMEM, 0,
                             "allocating delta stream",
                             NULL, pool);

  (*stream)->source_fn = source_fn; 
  (*stream)->source_data = source_data;
  (*stream)->target_fn = target_fn;
  (*stream)->target_data = target_data;
  (*stream)->pool = pool;
  (*stream)->more = TRUE;
  return NULL;
}


void
svn_free_delta_stream (svn_delta_stream_t* stream)
{
  if (stream)
    apr_destroy_pool (stream->pool);
}



svn_error_t*
svn_next_delta_window (svn_delta_stream_t* stream,
                       svn_delta_window_t** window)
{
  if (!stream->more)
    {
      *window = NULL;
      return NULL;
    }
  else
    {
      return svn_tdelta__vdelta (window, stream);
    }
}


void
svn_free_delta_window (svn_delta_window_t* window)
{
  if (window)
    apr_destroy_pool (window->pool);
}



/* 
 * local variables:
 * eval: (load-file "../svn-dev.el")
 * end:
 */
